
/*
 * Experimental PJSIP channel module for FreeSWITCH
 *
 * 
 *
 *
 *
 *
 */
/***** ChangeLog
 *
 *  2006-06-18: initial version.
 *              99% copy'n'paste from FreeSWITCH and pjsip example code.
 *
 *  2006-06-20: added parts of sdp negotiation, started writing pjsip event handler code.
 *              more code taken from mod_exosip for FS event handling.
 *
 *  2006-07-12: initial version of profiles support
 *
 */
/*
 * TODO:
 *
 * Do we really need to call pj_register_thread() in monitor_thread()?
 *
 */

#define HAVE_APR
#include <switch.h>
#include <pjsip.h>
#include <pjsdp.h>
#include <pjsip_ua.h>
#include <pjsip_simple.h>
#include <pjlib-util.h>
#include <pjlib.h>

#include "mod_pjsip.h"

static const char modname[] = "mod_pjsip";

#define MOD_PJSIP_CFG	"pjsip.conf"

static switch_memory_pool_t *module_pool = NULL;
static pj_caching_pool cp;
static pjsip_inv_callback invite_cb;
static pj_pool_t *pjsip_pool;
static pj_str_t FREESWITCH_SERVER;

#define DEFAULT_PORT 5060

#define PJ_MAX_STRLEN(str, pjstr) ( sizeof(str) > pjstr.slen ? pjstr.slen : sizeof(str) )

static struct {
	/* */
	switch_hash_t *profiles_hash;
	switch_hash_t *transport_hash;
	int running;

	pjsip_endpoint *sip_endpt;
	pjsip_module mod_app;
} globals;

struct pjsip_profile {
	pj_str_t local_uri;
	pj_str_t local_contact;

	pj_sockaddr_in sip_addr;
	char *sip_ip;
	int sip_port;

	/* */
	char name[128];
	char local_addr[128];
	char *dialplan;
	char *codec_string;
	char *codec_order[SWITCH_MAX_CODECS];
	int codec_order_last;

	int max_calls;
	int calls;

	int flags;
	switch_hash_t *call_hash;
};

typedef enum {
	TFLAG_OUTBOUND = (1 << 0),
	TFLAG_INBOUND = (1 << 1),
	TFLAG_IO = (1 << 2),
	TFLAG_BYE = (1 << 3),
	TFLAG_ANSWERED = (1 << 4),
	TFLAG_USING_CODEC = (1 << 5),
	TFLAG_RTP = (1 << 6),

	TFLAG_READING = (1 << 10),
	TFLAG_WRITING = (1 << 11),
	TFLAG_TIMER = (1 << 20)
} tech_flag_t;

typedef enum {
	PFLAG_TRANSPORT_UDP = (1 << 0),
	PFLAG_TRANSPORT_TCP = (1 << 1),
	PFLAG_TRANSPORT_TLS = (1 << 2)
} profile_flag_t;

#define MAX_TRANSPORTS 3

#define set_param(ptr, val) if ( ptr ) { free( ptr ); ptr = NULL; } if ( val ) { ptr = strdup( val ); }

static switch_status_t create_sdp( pj_pool_t *pool,
					struct private_object *tech_pvt,
					pjmedia_sdp_session **p_sdp,
					pjmedia_sdp_session *ref_sdp );

static switch_status_t start_rtp( struct private_object *tech_pvt );
static switch_status_t stop_rtp( struct private_object *tech_pvt );
//static struct private_object * find_pvt_by_callid( char *id );

/* BEGIN: freeswitch callbacks */
/* event handlers */
static switch_status_t pjsip_on_init(switch_core_session_t *session);
static switch_status_t pjsip_on_hangup(switch_core_session_t *session);
static switch_status_t pjsip_on_loopback(switch_core_session_t *session);
static switch_status_t pjsip_on_transmit(switch_core_session_t *session); 

/* io routines */
static switch_status_t pjsip_outgoing_channel(switch_core_session_t *session, switch_caller_profile_t *outbound_profile,
                                             switch_core_session_t **new_session, switch_memory_pool_t *pool);

static switch_status_t pjsip_read_frame(switch_core_session_t *session, switch_frame_t **frame, int timeout,
                                             switch_io_flag_t flags, int stream_id);

static switch_status_t pjsip_write_frame(switch_core_session_t *session, switch_frame_t *frame, int timeout,
                                             switch_io_flag_t flags, int stream_id);

static switch_status_t pjsip_answer_channel(switch_core_session_t *session);
static switch_status_t pjsip_kill_channel(switch_core_session_t *session, int sig);	

static switch_status_t pjsip_waitfor_read(switch_core_session_t *session, int ms, int stream_id);
static switch_status_t pjsip_waitfor_write(switch_core_session_t *session, int ms, int stream_id);

static switch_status_t pjsip_on_ring(switch_core_session_t *session);

static switch_status_t pjsip_receive_message(switch_core_session_t *session,
						switch_core_session_message_t *msg);

static const switch_io_routines_t pjsip_io_routines = {
        /*.outgoing_channel */ &pjsip_outgoing_channel,
        /*.answer_channel */ &pjsip_answer_channel,
        /*.read_frame */ &pjsip_read_frame,
        /*.write_frame */ &pjsip_write_frame,
        /*.kill_channel */ &pjsip_kill_channel,
        /*.waitfor_read */ &pjsip_waitfor_read,
        /*.waitfor_read */ &pjsip_waitfor_write,
        /*.send_dtmf*/ NULL,
        /*.receive_message*/ &pjsip_receive_message
};

static const switch_state_handler_table_t pjsip_event_handlers = {
        /*.on_init */ &pjsip_on_init,
        /*.on_ring */ &pjsip_on_ring,
        /*.on_execute */ NULL,
        /*.on_hangup */ &pjsip_on_hangup,
        /*.on_loopback */ &pjsip_on_loopback,
        /*.on_transmit */ &pjsip_on_transmit
};

static const switch_endpoint_interface_t pjsip_endpoint_interface = {
        /*.interface_name */ "pjsip",
        /*.io_routines */ &pjsip_io_routines,
        /*.event_handlers */ &pjsip_event_handlers,
        /*.private */ NULL,
        /*.next */ NULL
};

static const switch_loadable_module_interface_t pjsip_module_interface = {
        /*.module_name */ modname,
        /*.endpoint_interface */ &pjsip_endpoint_interface,
        /*.timer_interface */ NULL,
        /*.dialplan_interface */ NULL,
        /*.codec_interface */ NULL,
        /*.application_interface */ NULL
};


void log_func(int level, const char *data, int len)
{
	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, (char *) data);
}


/* */
static int pjsip_strncpy( char *dst, const pj_str_t *src, int maxlen )
{
	int len = maxlen > src->slen ? src->slen : maxlen;

	switch_copy_string( dst, src->ptr, len - 1 );
	dst[len] = '\0';

	return len;
}


static void add_server_header(pjsip_tx_data *tdata)
{
	const pj_str_t SERVER = { "Server", 6};
	pjsip_hdr *h;
	
	h = (pjsip_hdr*) pjsip_generic_string_hdr_create(tdata->pool,
													 &SERVER,
													 &FREESWITCH_SERVER);
	pjsip_msg_add_hdr(tdata->msg, h);

}



static switch_status_t pjsip_on_init(switch_core_session_t *session)
{
	struct pjsip_profile *profile;
	struct private_object *tech_pvt;
	switch_channel_t *channel;
	pj_status_t status;

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	profile = tech_pvt->profile;
	assert( profile != NULL );

	/* register switch thread to pjsip core */
	status = pj_thread_register( NULL, tech_pvt->thread_desc, &tech_pvt->thread );
	if( status != PJ_SUCCESS ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Could not register switch session thread to pjlib\n" );
		switch_core_session_destroy( &session );
		return SWITCH_STATUS_GENERR;
	}

	if( switch_channel_test_flag( channel, CF_OUTBOUND ) )
	{
		pjsip_tx_data *txdata = NULL;
		pjmedia_sdp_session *sdp = NULL;
		char tmp[1024], *destination;
		pj_str_t dst;

		switch_copy_string( tmp, tech_pvt->caller_profile->destination_number, sizeof( tmp ) );
		if( !(destination = strchr( tmp, '/' )) ) {
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Malformed URL!\n" );
			switch_core_session_destroy( &session );
			return SWITCH_STATUS_GENERR;
		}
		*destination++ = '\0';

		pj_ansi_sprintf( tmp, "<sip:%s:%d>", destination, 5060 );
		pj_cstr( &dst, tmp );

		tech_pvt->sip_ua = pjsip_ua_instance();

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Dst: %s, Local: %s %s\n", tmp, tech_pvt->profile->local_uri.ptr, tech_pvt->profile->local_contact.ptr );

		/* create new UAC dialog */
		status = pjsip_dlg_create_uac( tech_pvt->sip_ua,
						&tech_pvt->profile->local_uri,
						&tech_pvt->profile->local_contact,
						&dst,
						&dst,
						&tech_pvt->sip_dialog );

		if ( status != PJ_SUCCESS ) {
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new UAC dialog for call\n" );
			switch_core_session_destroy( &session );
			return SWITCH_STATUS_GENERR;
		}

		/* create SDP */
		create_sdp( tech_pvt->sip_dialog->pool, tech_pvt, &sdp, NULL );

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Initial SDP created\n" );

		/* create INVITE session */
		if ((status = pjsip_inv_create_uac( tech_pvt->sip_dialog, sdp, 0, &tech_pvt->sip_invite ) != PJ_SUCCESS )) {
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new INVITE session for call %s\n" );
			pjsip_dlg_terminate( tech_pvt->sip_dialog );
			return SWITCH_STATUS_GENERR;
		}


		/* attach private object to pjsip session data */
		tech_pvt->sip_invite->mod_data[globals.mod_app.id] = tech_pvt;

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Invite UAC created\n" );

		/* create invite request */
		status = pjsip_inv_invite( tech_pvt->sip_invite, &txdata );
		if ( status != PJ_SUCCESS ) {
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create initial INVITE request for call %s\n" );
			pjsip_inv_terminate(tech_pvt->sip_invite, 500, PJ_FALSE);
			return SWITCH_STATUS_GENERR;
		}

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Initial INVITE created\n" );

		/* set call id (now, before invite is sent, so that when callback is called,
		 * tech_pvt has the correct call_id) 
		 */
		strncpy( tech_pvt->call_id, tech_pvt->sip_invite->dlg->call_id->id.ptr, PJ_MAX_STRLEN( tech_pvt->call_id, tech_pvt->sip_invite->dlg->call_id->id ) );

		/* send request */
		add_server_header(txdata);
		status = pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );
		if ( status != PJ_SUCCESS ) {
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to send INVITE request for call %s\n" );
			pjsip_inv_terminate(tech_pvt->sip_invite, 500, PJ_FALSE);
			return SWITCH_STATUS_GENERR;
		}

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "INVITE sent\n" );

	}

	
	/* add call to hash */
	switch_core_hash_insert( profile->call_hash, tech_pvt->call_id, tech_pvt );

	switch_channel_set_state( channel, CS_RING );
	return SWITCH_STATUS_SUCCESS;
} 

static switch_status_t pjsip_on_ring(switch_core_session_t *session) 
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;

	assert( session != NULL );

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RING %s\n", switch_channel_get_name( channel ) );

	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t pjsip_on_hangup(switch_core_session_t *session)
{ 
	struct pjsip_profile *profile;
	struct private_object *tech_pvt;
	switch_channel_t *channel;
	pj_status_t status;
	pjsip_tx_data *txdata;

	assert( session != NULL );

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	profile = tech_pvt->profile;
	assert( profile != NULL );

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Hangup call %s\n", tech_pvt->call_id );

	/* remove call from hash */
	switch_core_hash_delete( profile->call_hash, tech_pvt->call_id );

	switch_set_flag_locked( tech_pvt, TFLAG_BYE );
	switch_clear_flag_locked( tech_pvt, TFLAG_IO );

	stop_rtp( tech_pvt );

	if( tech_pvt->sip_invite->state < PJSIP_INV_STATE_CONFIRMED ) {
		pjsip_inv_terminate( tech_pvt->sip_invite, 500, PJ_TRUE );
	} else {
		status = pjsip_inv_end_session( tech_pvt->sip_invite, 200, NULL, &txdata );
		if( status != PJ_SUCCESS )
			pjsip_inv_terminate( tech_pvt->sip_invite, 500, PJ_TRUE );
		else {
			add_server_header(txdata);
			pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );
		}
	}

	if ( switch_test_flag( tech_pvt, TFLAG_USING_CODEC ) ) {
		switch_core_codec_destroy( &tech_pvt->read_codec );
		switch_core_codec_destroy( &tech_pvt->write_codec );
	}
	return SWITCH_STATUS_SUCCESS;
} 

static switch_status_t pjsip_on_loopback(switch_core_session_t *session)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	/* do nothing */

	return SWITCH_STATUS_SUCCESS;
} 

static switch_status_t pjsip_on_transmit(switch_core_session_t *session)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;

	assert( session != NULL );

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	/* do nothing */

	return SWITCH_STATUS_SUCCESS;
} 

/* answer channel */
static switch_status_t pjsip_answer_channel(switch_core_session_t *session)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;

	assert( session != NULL );

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Answering call %s\n", tech_pvt->call_id );

	if( !switch_test_flag( tech_pvt, TFLAG_ANSWERED ) &&
	    !switch_channel_test_flag( channel, CF_OUTBOUND ) )
	{
		pjsip_tx_data *txdata;
		pj_status_t status;

		assert( tech_pvt->sip_invite != NULL );

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Sending 200 on INVITE to %s (%s)\n", tech_pvt->call_id, switch_channel_get_name( channel ) );

		/* create SIP 200 with SDP */
		status = pjsip_inv_answer( tech_pvt->sip_invite, 200, NULL, NULL, &txdata );
		if( status != PJ_SUCCESS ) {
			pjsip_inv_terminate( tech_pvt->sip_invite, 500, PJ_TRUE );
			switch_core_session_destroy( &session );
			return SWITCH_STATUS_GENERR;
		}
		add_server_header(txdata);
		status = pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );
		if( status != PJ_SUCCESS ) {
			pjsip_inv_terminate( tech_pvt->sip_invite, 500, PJ_TRUE );
			switch_core_session_destroy( &session );
			return SWITCH_STATUS_GENERR;
		}
	}
	switch_set_flag_locked( tech_pvt, TFLAG_ANSWERED );

	return SWITCH_STATUS_SUCCESS;
}

/* kill channel */
static switch_status_t pjsip_kill_channel(switch_core_session_t *session, int sig)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;

	assert( session != NULL );

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	switch_mutex_lock(tech_pvt->flag_mutex);
	switch_clear_flag( tech_pvt, TFLAG_IO );
	switch_set_flag( tech_pvt, TFLAG_BYE );
	if( tech_pvt->rtp_session ) {
		switch_rtp_kill_socket( tech_pvt->rtp_session );
	}
	switch_mutex_unlock(tech_pvt->flag_mutex);

	return SWITCH_STATUS_SUCCESS;
}

/* wait for read */
static switch_status_t pjsip_waitfor_read(switch_core_session_t *session, int ms, int stream_id)
{
	/* do nothing */

	return SWITCH_STATUS_SUCCESS;
}

/* wait for write */
static switch_status_t pjsip_waitfor_write(switch_core_session_t *session, int ms, int stream_id)
{
	/* do nothing */

	return SWITCH_STATUS_SUCCESS;
}


/* create new call */
static switch_status_t pjsip_outgoing_channel(switch_core_session_t *session, switch_caller_profile_t *outbound_profile,
                                             switch_core_session_t **new_session, switch_memory_pool_t *pool)
{ 
	struct private_object *tech_pvt = NULL;
	switch_channel_t *channel = NULL;
	struct pjsip_profile *profile = NULL;
	switch_caller_profile_t *caller_profile = NULL;
	char tmp[1024], *profile_name, *dest;

	assert( outbound_profile != NULL );

	switch_copy_string( tmp, outbound_profile->destination_number, sizeof(tmp) );
	profile_name = tmp;
	if(( dest = strchr( tmp, '/' ) )) {
		*dest++ = '\0';
	}

	if( !( profile_name && dest ) ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Malformed URL!\n" );
		return SWITCH_STATUS_GENERR;
	}

	if( !( profile = (struct pjsip_profile *) switch_core_hash_find( globals.profiles_hash, profile_name ) )) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Unknown profile: %s\n", profile_name );
		return SWITCH_STATUS_GENERR;
	}

	if ( !( *new_session = switch_core_session_request( &pjsip_endpoint_interface, pool ) ) ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new session for call\n" );
		return SWITCH_STATUS_GENERR;
	}
	switch_core_session_add_stream( *new_session, NULL );

	/* create new call */
	if (!(tech_pvt = (struct private_object *) switch_core_session_alloc(*new_session, sizeof(struct private_object)))) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new private object for call\n" );
		goto err;
	}
	tech_pvt->session = *new_session;
	switch_mutex_init(&tech_pvt->flag_mutex, SWITCH_MUTEX_NESTED, switch_core_session_get_pool(*new_session));
	tech_pvt->profile = profile;

	switch_core_session_set_private( *new_session, tech_pvt );
	channel = switch_core_session_get_channel( *new_session );

	/* create channel name, copy profile and parse destination number (?) */
	snprintf( tmp, sizeof(tmp), "pjsip/%s-%04x", outbound_profile->destination_number, rand() & 0xffff );
	switch_channel_set_name( channel, tmp );

	/* get list of all available codecs */
	tech_pvt->num_codecs = switch_loadable_module_get_codecs( switch_core_session_get_pool(tech_pvt->session), tech_pvt->codecs,
								sizeof(tech_pvt->codecs) / sizeof(tech_pvt->codecs[0]) );

	caller_profile = switch_caller_profile_clone( *new_session, outbound_profile );
	switch_channel_set_caller_profile( channel, caller_profile );

	tech_pvt->caller_profile = caller_profile;

	switch_channel_set_flag( channel, CF_OUTBOUND );
	switch_channel_set_state( channel, CS_INIT );

	return SWITCH_STATUS_SUCCESS;

err:
	switch_core_session_destroy(new_session);
	return SWITCH_STATUS_GENERR;
} 


static switch_status_t pjsip_read_frame(switch_core_session_t *session, switch_frame_t **frame, int timeout,
									   switch_io_flag_t flags, int stream_id)
{
	struct private_object *tech_pvt = NULL;
	size_t bytes = 0, samples = 0, frames = 0, ms = 0;
	switch_channel_t *channel = NULL;
	int payload = 0;
	switch_time_t now, started = switch_time_now(), last_act = switch_time_now();
	unsigned int elapsed;
	uint32_t hard_timeout = 60000 * 3;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	tech_pvt->read_frame.datalen = 0;
	switch_set_flag_locked(tech_pvt, TFLAG_READING);


	samples = tech_pvt->read_codec.implementation->samples_per_frame;
	ms = tech_pvt->read_codec.implementation->microseconds_per_frame;
	
	if (tech_pvt->last_read) {
		elapsed = (unsigned int)((switch_time_now() - tech_pvt->last_read) / 1000);
		if (elapsed > 60000) {
			return SWITCH_STATUS_TIMEOUT;
		}
	}

	if (switch_test_flag(tech_pvt, TFLAG_IO)) {
		switch_status_t status;

		assert(tech_pvt->rtp_session != NULL);
		tech_pvt->read_frame.datalen = 0;


		while (!switch_test_flag(tech_pvt, TFLAG_BYE) && switch_test_flag(tech_pvt, TFLAG_IO) && tech_pvt->read_frame.datalen == 0) {
			now = switch_time_now();
			tech_pvt->read_frame.flags = 0;

			status = switch_rtp_zerocopy_read_frame(tech_pvt->rtp_session, &tech_pvt->read_frame);
			if (status != SWITCH_STATUS_SUCCESS && status != SWITCH_STATUS_BREAK) {
				return SWITCH_STATUS_FALSE;
			}
			
			payload = tech_pvt->read_frame.payload;


			elapsed = (unsigned int)((switch_time_now() - started) / 1000);

			if (timeout > -1) {
				if (elapsed >= (unsigned int)timeout) {
					return SWITCH_STATUS_BREAK;
				}
			}
			
			elapsed = (unsigned int)((switch_time_now() - last_act) / 1000);

			if (elapsed >= hard_timeout) {
				return SWITCH_STATUS_BREAK;
			}

			if (tech_pvt->read_codec.implementation->encoded_bytes_per_frame) {
				bytes	= tech_pvt->read_codec.implementation->encoded_bytes_per_frame;
				frames	= tech_pvt->read_frame.datalen / bytes;
			} else {
				frames = 1;
			}

			samples	= frames * tech_pvt->read_codec.implementation->samples_per_frame;


			if (switch_rtp_has_dtmf(tech_pvt->rtp_session)) {
				char dtmf[128];
				switch_rtp_dequeue_dtmf(tech_pvt->rtp_session, dtmf, sizeof(dtmf));
				switch_channel_queue_dtmf(channel, dtmf);
			}


			if (tech_pvt->read_frame.datalen > 0) {
				tech_pvt->last_read = switch_time_now();
				if (tech_pvt->read_codec.implementation->encoded_bytes_per_frame) {
					bytes = tech_pvt->read_codec.implementation->encoded_bytes_per_frame;
					frames = (tech_pvt->read_frame.datalen / bytes);
				} else {
					frames = 1;
				}
				samples = frames * tech_pvt->read_codec.implementation->samples_per_frame;
				ms = frames * tech_pvt->read_codec.implementation->microseconds_per_frame;
				tech_pvt->read_frame.samples = (int) samples;
				break;
			}

			switch_yield(1000);
		}

	}


	switch_clear_flag_locked(tech_pvt, TFLAG_READING);

	if (switch_test_flag(tech_pvt, TFLAG_BYE)) {
		switch_channel_hangup(channel, SWITCH_CAUSE_NORMAL_CLEARING);
		return SWITCH_STATUS_FALSE;
	}

	*frame = &tech_pvt->read_frame;

	return SWITCH_STATUS_SUCCESS;
}


static switch_status_t pjsip_write_frame(switch_core_session_t *session, switch_frame_t *frame, int timeout,
										switch_io_flag_t flags, int stream_id)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel = NULL;
	switch_status_t status = SWITCH_STATUS_SUCCESS;
	int bytes = 0, samples = 0, frames = 0;

	channel = switch_core_session_get_channel(session);
	assert(channel != NULL);

	tech_pvt = switch_core_session_get_private(session);
	assert(tech_pvt != NULL);

	if (!switch_test_flag(tech_pvt, TFLAG_IO)) {
		return SWITCH_STATUS_SUCCESS;
	}

	if (switch_test_flag(tech_pvt, TFLAG_BYE)) {
		switch_channel_hangup(channel, SWITCH_CAUSE_NORMAL_CLEARING);
		return SWITCH_STATUS_FALSE;
	}

	switch_set_flag_locked(tech_pvt, TFLAG_WRITING);



	if (tech_pvt->read_codec.implementation->encoded_bytes_per_frame) {
		bytes = tech_pvt->read_codec.implementation->encoded_bytes_per_frame;
		frames = ((int) frame->datalen / bytes);
	} else {
		frames = 1;
	}

	samples	= frames * tech_pvt->read_codec.implementation->samples_per_frame;
	switch_rtp_write_frame( tech_pvt->rtp_session, frame, samples );

	switch_clear_flag_locked(tech_pvt, TFLAG_WRITING);
	return status;
}


/* END: freeswitch callbacks */



/* BEGIN: pjsip callbacks */

/*
 *
 */
static void on_incoming_call( pjsip_rx_data *rxdata )
{
	switch_core_session_t *session;
	struct pjsip_profile *profile;
	pj_status_t status;

	/* How to map rxdata / transport -> profile? */
	if( !( profile = switch_core_hash_find( globals.transport_hash, rxdata->tp_info.transport->obj_name )) ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Profile not found\n" );
		return;
	}

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "*** New incoming call for profile %s\n", profile->name );

	/* check call limit */
	if( profile->max_calls && ( profile->calls + 1 ) > profile->max_calls ) {
		const pj_str_t reason = pj_str("Too many calls");
		pjsip_endpt_respond_stateless( globals.sip_endpt, rxdata, 
						500, &reason,
						NULL, NULL);

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Too many calls (profile: %s)\n", profile->name );
		return;
	}

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "PJSIP process id %lu, thread name %s\n", pj_getpid(), pj_thread_get_name( pj_thread_this() ) );

	if ((session = switch_core_session_request( &pjsip_endpoint_interface, NULL ))) {
		struct private_object *tech_pvt;
		switch_channel_t *channel;
		pjmedia_sdp_session *sdp;
		pjsip_dialog *dialog;
		pjsip_tx_data *txdata;
		pj_str_t tmp, username, host, destination;
		char name[1024];

		channel = switch_core_session_get_channel( session );

		/* allocate new session */
		switch_core_session_add_stream( session, NULL );
		if (!(tech_pvt = (struct private_object *) switch_core_session_alloc( session, sizeof(struct private_object) ))) {
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new channel for inbound call\n" );
			switch_core_session_destroy( &session );
			return;
		}

		tech_pvt->session = session;
		switch_core_session_set_private( session, tech_pvt );


		switch_mutex_init(&tech_pvt->flag_mutex, SWITCH_MUTEX_NESTED, switch_core_session_get_pool(session));
		switch_mutex_lock(tech_pvt->flag_mutex);

		/* need the invite data in pjsip_on_init */
		tech_pvt->sip_ua = pjsip_ua_instance();
		tech_pvt->profile = profile;
		
		/* give the new session a name */
		pj_strdup_with_null( pjsip_pool, &tmp, &((pjsip_sip_uri *)rxdata->msg_info.from->uri)->host );
		snprintf( name, sizeof(name), "pjsip/%s/%s-%04x", profile->name, tmp.ptr, rand() & 0xffff );
		switch_channel_set_name( channel, name );

		switch_channel_clear_flag( channel, CF_OUTBOUND );

		/* create UAS dialog */
		status = pjsip_dlg_create_uas( tech_pvt->sip_ua, rxdata, NULL, &dialog );
		if( status != PJ_SUCCESS ) {
			const pj_str_t reason = pj_str("Unable to create dialog");
			pjsip_endpt_respond_stateless( globals.sip_endpt, rxdata, 
							500, &reason,
							NULL, NULL);
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new pjsip dialog for inbound call\n" );
			switch_mutex_unlock(tech_pvt->flag_mutex);
			switch_core_session_destroy( &session );
			return;
		}

		/* lock dialog, to prevent callbacks to be called while we're setting
		 * up the dialog
		 */
		pjsip_dlg_inc_lock(dialog);

		/* load codecs */
		tech_pvt->num_codecs = switch_loadable_module_get_codecs( switch_core_session_get_pool(tech_pvt->session), tech_pvt->codecs,
																  sizeof(tech_pvt->codecs) / sizeof(tech_pvt->codecs[0]) );
		
		/* Create our initial sdp */
		create_sdp( dialog->pool, tech_pvt, &sdp, NULL );
		tech_pvt->sip_dialog = dialog;

		
		/* Create UAS invite session */
		if ((status = pjsip_inv_create_uas( dialog, rxdata, sdp, 0, &tech_pvt->sip_invite ) != PJ_SUCCESS )) {
			pjsip_hdr hdr_list;
			pjsip_warning_hdr *w;

			/* Send 500, adding Warning header to explain why it went wrong */
			w = pjsip_warning_hdr_create_from_status(dialog->pool, 
								 pjsip_endpt_name(globals.sip_endpt),
								 status);
			pj_list_init(&hdr_list);
			pj_list_push_back(&hdr_list, w);

			pjsip_dlg_respond(dialog, rxdata, 500, NULL, &hdr_list, NULL);

			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new pjsip dialog for inbound call\n" );
			switch_core_session_destroy( &session );

			/* unlock (and probably destroy) dialog */
			pjsip_dlg_dec_lock(dialog);
			switch_mutex_unlock(tech_pvt->flag_mutex);
			return;
		}

		
		tech_pvt->sip_invite->mod_data[globals.mod_app.id] = tech_pvt;
		tech_pvt->profile = profile;

		pj_strdup_with_null( pjsip_pool, &username, &((pjsip_sip_uri *)rxdata->msg_info.from->uri)->user );
		pj_strdup_with_null( pjsip_pool, &host, &((pjsip_sip_uri *)rxdata->msg_info.from->uri)->host );
		pj_strdup_with_null( pjsip_pool, &destination, &((pjsip_sip_uri *)rxdata->msg_info.to->uri)->host );

		if ( (tech_pvt->caller_profile = switch_caller_profile_new( switch_core_session_get_pool( session ),
													username.ptr,
													profile->dialplan,
													username.ptr,
													username.ptr,
													host.ptr,
													NULL,
													NULL,
													NULL,
													(char *)modname,
													NULL,
													destination.ptr)) != 0 ) {
			switch_channel_set_caller_profile( channel, tech_pvt->caller_profile );
		}
		
		/* set call id */
		pjsip_strncpy( tech_pvt->call_id, &dialog->call_id->id, sizeof( tech_pvt->call_id ) );

		switch_set_flag_locked( tech_pvt, TFLAG_INBOUND );

		/* Send 100 Trying */
		status = pjsip_inv_initial_answer( tech_pvt->sip_invite, rxdata, 100, NULL, NULL, &txdata );
		if( status != PJ_SUCCESS ) {
			pjsip_dlg_respond(dialog, rxdata, 500, NULL, NULL, NULL);
			pjsip_inv_terminate(tech_pvt->sip_invite, 500, PJ_FALSE);

			switch_core_session_destroy( &session );

			/* unlock (and probably destroy) dialog */
			pjsip_dlg_dec_lock(dialog);
			switch_mutex_unlock(tech_pvt->flag_mutex);
			return;
		}
		add_server_header(txdata);
		pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );

		/* Print the initial SDP to the console / log */
		{
			char tmp_sdp[1024];
			int len = pjmedia_sdp_print( sdp, tmp_sdp, sizeof(tmp_sdp)-1 );
			if( len > 0 ) {
				tmp_sdp[len] = '\0';
				switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, 
						"\n** Our initial SDP **\n%s\n*********************\n", tmp_sdp );
			}
		}
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "Starting session thread for call %s\n", tech_pvt->call_id );

		/* set initial state */
		switch_channel_set_state( channel, CS_INIT );

		/* Send 180 Ring */
		status = pjsip_inv_answer( tech_pvt->sip_invite, 183, NULL, NULL, &txdata );
		pjsip_tx_data_dec_ref(txdata);
		status = pjsip_inv_answer( tech_pvt->sip_invite, 180, NULL, NULL, &txdata );
		if( status != PJ_SUCCESS ) {
			status = pjsip_inv_answer( tech_pvt->sip_invite, PJSIP_SC_NOT_ACCEPTABLE, NULL, NULL, &txdata );

			if( status == PJ_SUCCESS )
				pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );
			else
				pjsip_inv_terminate( tech_pvt->sip_invite, 500, PJ_FALSE );

			switch_core_session_destroy( &session );
			return;
		}
		add_server_header(txdata);
		pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );

		switch_mutex_unlock(tech_pvt->flag_mutex);
		switch_core_session_thread_launch( session );

		/* Done setting up UAS invite session, release lock */
		pjsip_dlg_dec_lock(dialog);
		return;
	} else {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create new channel for inbound call\n" );
		return;
	}
}

/* Handle incoming OPTIONS request */
static void on_incoming_options( pjsip_rx_data *rdata )
{
    pjsip_tx_data *tdata;
    pjsip_response_addr res_addr;
    //pjmedia_sdp_session *sdp;
    const pjsip_hdr *cap_hdr;
    pj_status_t status;
	
    /* Create basic response. */
    status = pjsip_endpt_create_response(globals.sip_endpt, rdata, 200, NULL, &tdata);
    if (status != PJ_SUCCESS) {
		return;
    }
	
    /* Add Allow header */
    cap_hdr = pjsip_endpt_get_capability(globals.sip_endpt, PJSIP_H_ALLOW, NULL);
    if (cap_hdr) {
		pjsip_msg_add_hdr(tdata->msg, pjsip_hdr_clone(tdata->pool, cap_hdr));
    }
	
    /* Add Accept header */
    cap_hdr = pjsip_endpt_get_capability(globals.sip_endpt, PJSIP_H_ACCEPT, NULL);
    if (cap_hdr) {
		pjsip_msg_add_hdr(tdata->msg, pjsip_hdr_clone(tdata->pool, cap_hdr));
    }
	
    /* Add Supported header */
    cap_hdr = pjsip_endpt_get_capability(globals.sip_endpt, PJSIP_H_SUPPORTED, NULL);
    if (cap_hdr) {
		pjsip_msg_add_hdr(tdata->msg, pjsip_hdr_clone(tdata->pool, cap_hdr));
    }
	
    /* Add Allow-Events header from the evsub module */
    cap_hdr = pjsip_evsub_get_allow_events_hdr(NULL);
    if (cap_hdr) {
		pjsip_msg_add_hdr(tdata->msg, pjsip_hdr_clone(tdata->pool, cap_hdr));
    }
	
    /* Add Server header */
	add_server_header(tdata);
	
    /* TODO: it would be nice if we could add SDP body here (dunno how) */
	//PJ_TODO(ADD_SDP_TO_OPTIONS_RESPONSE);
#if 0
    status = pjmedia_endpt_create_sdp(pjsua_var.med_endpt, tdata->pool, 1,
		&pjsua_var.calls[0].skinfo, &sdp);
    if (status == PJ_SUCCESS) {
		pjsip_create_sdp_body(tdata->pool, sdp, &tdata->msg->body);
    }
#endif
	
    /* Send response statelessly */
    pjsip_get_response_addr(tdata->pool, rdata, &res_addr);
    status = pjsip_endpt_send_response(globals.sip_endpt, &res_addr, tdata, NULL, NULL);
    if (status != PJ_SUCCESS)
		pjsip_tx_data_dec_ref(tdata);
}


static pj_bool_t on_rx_request( pjsip_rx_data *rxdata )
{ 
	pj_str_t reason;

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "%s: received new request\n", __FUNCTION__ );

	switch( rxdata->msg_info.msg->line.req.method.id ) {
		case PJSIP_INVITE_METHOD:
			on_incoming_call( rxdata );
			return PJ_TRUE;
			;;
		case PJSIP_OPTIONS_METHOD:
			on_incoming_options( rxdata );
			return PJ_TRUE;
			;;
		default:
			reason = rxdata->msg_info.msg->line.req.method.name;
			pjsip_endpt_respond_stateless( rxdata->tp_info.transport->endpt, rxdata, 
							500, &reason, NULL, NULL );
			;;
	}
	return PJ_TRUE;
}

static void call_on_tsx_state_changed( pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e )
{
	struct private_object *tech_pvt;
	struct pjsip_profile *profile;
	switch_core_session_t *session;

	assert( inv != NULL && tsx != NULL );


	if (!(tech_pvt = inv->mod_data[globals.mod_app.id])) {
		return;
	}

	switch_mutex_lock(tech_pvt->flag_mutex);

	profile = tech_pvt->profile;
	assert( profile != NULL );

	session = tech_pvt->session;
	assert( session != NULL );

	switch_mutex_unlock(tech_pvt->flag_mutex);

	if( tsx->method.id == PJSIP_INVITE_METHOD &&
		inv->state == PJSIP_INV_STATE_EARLY &&
	    tsx->state == PJSIP_TSX_STATE_PROCEEDING && 
	    tsx->status_code == PJSIP_SC_RINGING )
	{
		switch_channel_t *channel;

		channel = switch_core_session_get_channel( session );
		assert( channel != NULL );

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Ring received (call %s)\n", tech_pvt->call_id );
	}
}

static void call_on_forked( pjsip_inv_session *inv,
					pjsip_event *evt ) { }

/* pjsip callback for inv session state changes */
static void call_on_state_changed( pjsip_inv_session *inv,
					pjsip_event *evt )
{ 
	struct private_object *tech_pvt;
	struct pjsip_profile *profile;
	switch_core_session_t *session;
	switch_channel_t *channel;

	assert ( evt != NULL);
	assert ( inv != NULL );
	
	if (!(tech_pvt = inv->mod_data[globals.mod_app.id])) {
		return;
	}

	assert( tech_pvt->flag_mutex != NULL );

	switch_mutex_lock(tech_pvt->flag_mutex);

	session = tech_pvt->session;
	assert ( session != NULL );

	profile = tech_pvt->profile;
	assert( profile != NULL );

	switch_mutex_unlock(tech_pvt->flag_mutex);


	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );


	
	switch( inv->state )
	{
	case PJSIP_INV_STATE_DISCONNECTED:
		inv->mod_data[globals.mod_app.id] = NULL;
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "* Call %s (%s) has been disconnected\n", tech_pvt->call_id,
									switch_channel_get_name( channel ) );

		if( !switch_test_flag( tech_pvt, TFLAG_BYE ) ) {
			switch_call_cause_t cause;

			switch( evt->body.rx_msg.rdata->msg_info.msg->line.status.code )
			{
			default:
				/* bennylp: we can get the INVITE disconnect cause from inv->cause.
				 *          and the reason in inv->cause_text.
				 *          But how do we map it to switch cause?
				 */
				//PJ_TODO(MAP_SIP_STATUS_CODE_TO_SWITCH_CAUSE_CODE);
				cause = SWITCH_CAUSE_NORMAL_CLEARING;
			}
			switch_channel_hangup( channel, cause );
					}
		break;
	case PJSIP_INV_STATE_CONFIRMED:
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "* Call %s (%s) has been connected\n", tech_pvt->call_id,
									switch_channel_get_name( channel ) );

		/* -> connected */
		if( switch_channel_test_flag( channel, CF_OUTBOUND ) ) {
			switch_channel_answer( channel );
		}

		break;
	case PJSIP_INV_STATE_EARLY:
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Call %s (%s), state %s\n", tech_pvt->call_id, 
									switch_channel_get_name( channel ),
									pjsip_inv_state_name( inv->state ) );

#if 0
		/* -> inbound ringing */
		if( !switch_channel_test_flag( channel, CF_OUTBOUND ) &&
		    !( switch_channel_get_state( channel ) >= CS_RING ) ) {

			/* put inbound channel into RING state */
			switch_channel_set_state( channel, CS_RING );
		}
#endif
		break;

	case PJSIP_INV_STATE_CONNECTING:
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Call %s (%s), state %s\n", tech_pvt->call_id,
									switch_channel_get_name( channel ),
									pjsip_inv_state_name( inv->state ) );

#if 0
		/* -> outbound ringing */
		if( switch_channel_test_flag( channel, CF_OUTBOUND ) &&
		    !( switch_channel_get_state( channel ) >= CS_RING ) ) {

			switch_channel_set_state( channel, CS_RING );
		}
#endif
		break;
	case PJSIP_INV_STATE_CALLING:
	case PJSIP_INV_STATE_INCOMING:
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Call %s (%s), state %s\n", tech_pvt->call_id,
									switch_channel_get_name( channel ),
									pjsip_inv_state_name( inv->state ) );
		break;
	default:
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "* Call %s (%s), UNHANDLED state %s\n", tech_pvt->call_id,
									switch_channel_get_name( channel ),
									pjsip_inv_state_name( inv->state ) );
		break;
	}
} 

/* pjsip callback for sdp negotiation changes */
static void call_on_media_update( pjsip_inv_session *inv,
					pj_status_t status )
{ 
	struct private_object *tech_pvt;
	struct pjsip_profile *profile;
	const pjmedia_sdp_session *remote_sdp, *local_sdp;
	pj_pool_t *pool;
	const switch_codec_implementation_t *codec, *imp;
	switch_channel_t *channel;
	switch_core_session_t *session;
	switch_status_t ret;
	int ms, rate;
	pj_sockaddr addr;

	tech_pvt = inv->mod_data[globals.mod_app.id];
	assert( tech_pvt != NULL );

	switch_mutex_lock(tech_pvt->flag_mutex);
	profile = tech_pvt->profile;
	assert( profile != NULL );
	switch_mutex_unlock(tech_pvt->flag_mutex);

	pool = inv->dlg->pool;
	assert( pool != NULL );

	session = tech_pvt->session;
	assert( session != NULL );

	channel = switch_core_session_get_channel( tech_pvt->session );
	assert( channel != NULL );

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "-- SDP negotiation status change %s (state %s, status %d)\n", tech_pvt->call_id,
					pjmedia_sdp_neg_state_str(pjmedia_sdp_neg_get_state( inv->neg )), status );

	if( pjmedia_sdp_neg_get_state( inv->neg ) != PJMEDIA_SDP_NEG_STATE_DONE ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "---- SDP negotiation still ongoing for call %s\n", tech_pvt->call_id );
	}

	if( status != PJ_SUCCESS ) {
		char tmp[1024];
		pj_strerror( status, tmp, sizeof(tmp) );
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "SDP negotiation failed for call %s, reason: %s\n", tech_pvt->call_id, tmp );
		switch_channel_hangup( channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER );
		return;
	}

	if( tech_pvt->rtp_session ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "SDP re-negotiation received %s\n", tech_pvt->call_id );

		stop_rtp( tech_pvt );
		switch_clear_flag_locked( tech_pvt, TFLAG_IO );

		/* destroy codecs */
		if ( switch_test_flag( tech_pvt, TFLAG_USING_CODEC ) ) {
			switch_core_codec_destroy( &tech_pvt->read_codec );
			switch_core_codec_destroy( &tech_pvt->write_codec );
		}
	}

	pjmedia_sdp_neg_get_active_local( inv->neg, &local_sdp );
	pjmedia_sdp_neg_get_active_remote( inv->neg, &remote_sdp );

	/* Print the final(?) SDPs to the console / log */
	{
		char tmp_sdp[1024];
		int len;

		len = pjmedia_sdp_print( local_sdp, tmp_sdp, sizeof(tmp_sdp)-1 );
		if( len > 0 ) {
			tmp_sdp[len] = '\0';
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, 
					"\n** Our final SDP **\n%s\n*********************\n", tmp_sdp );
		}

		len = pjmedia_sdp_print( remote_sdp, tmp_sdp, sizeof(tmp_sdp)-1 );
		if( len > 0 ) {
			tmp_sdp[len] = '\0';
			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, 
					"\n** Remote final SDP **\n%s\n*************************\n", tmp_sdp );
		}
	}

	/* Check if call is put on hold. */
	if (local_sdp->media_count == 0 ||
		pjmedia_sdp_media_find_attr2(local_sdp->media[0], "inactive", NULL) != NULL)
	{
		//PJ_TODO(HANDLE_CALL_HOLD);
	}


	/* retrieve stream (= codec) information */
	status = mod_pjsip_codec_from_sdp( tech_pvt, pool, local_sdp, remote_sdp, 0, &codec, &imp );
	if( status != PJ_SUCCESS ) {
		/* No valid codec found */
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "SDP negotation failed, codec unknown\n" );
		switch_set_flag_locked( tech_pvt, TFLAG_BYE );
		switch_clear_flag_locked( tech_pvt, TFLAG_IO );
		switch_channel_hangup( channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER );
		return;
	}

	/* valid codec + implementation found, setup codecs */
	rate = imp->bits_per_second / 8;
	  ms = imp->microseconds_per_frame / 1000;

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Codec %s, rate: %d, ms per frame: %d\n", codec->iananame, rate, ms );

	/* initialize codecs */
	ret = switch_core_codec_init(&tech_pvt->read_codec, codec->iananame, rate, ms, 1,
				SWITCH_CODEC_FLAG_ENCODE | SWITCH_CODEC_FLAG_DECODE,
				NULL, switch_core_session_get_pool( session ));

	if( ret != SWITCH_STATUS_SUCCESS ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize read codec\n" );
		switch_channel_hangup( channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER );
		return;
	}

	ret = switch_core_codec_init(&tech_pvt->write_codec, codec->iananame, rate, ms, 1,
				SWITCH_CODEC_FLAG_ENCODE | SWITCH_CODEC_FLAG_DECODE,
				NULL, switch_core_session_get_pool( session ));

	if( ret != SWITCH_STATUS_SUCCESS ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize write codec\n" );
		switch_channel_hangup( channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER );
		return;
	}

	tech_pvt->read_frame.rate  = rate;
	tech_pvt->read_frame.codec = &tech_pvt->read_codec;

	switch_core_session_set_read_codec( session, &tech_pvt->read_codec );
	switch_core_session_set_write_codec( session, &tech_pvt->write_codec );

	/* update remote rtp IP and port, restart rtp */
	status = mod_pjsip_addr_from_sdp( remote_sdp, 0, &addr );
	if( status != PJ_SUCCESS ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to get remote RTP endpoint address and port\n" );
		switch_channel_hangup( channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER );
	} else {
		pj_in_addr tmp_addr;

		tech_pvt->remote_sdp_audio_port = pj_sockaddr_in_get_port( (pj_sockaddr_in *)&addr );
		tmp_addr.s_addr = pj_htonl( pj_sockaddr_in_get_addr( (pj_sockaddr_in *)&addr ).s_addr );
		strncpy( tech_pvt->remote_sdp_audio_addr, pj_inet_ntoa( tmp_addr ), 
				sizeof(tech_pvt->remote_sdp_audio_addr) );
	}
	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Codec %s, %s:%d\n", codec->iananame, tech_pvt->remote_sdp_audio_addr, tech_pvt->remote_sdp_audio_port );

	/* start rtp */
	start_rtp( tech_pvt );

	/* activate io */
	switch_set_flag_locked( tech_pvt, TFLAG_IO );

}

static switch_status_t pjsip_receive_message(switch_core_session_t *session, switch_core_session_message_t *msg)
{
	struct private_object *tech_pvt;
	switch_channel_t *channel;

	tech_pvt = switch_core_session_get_private( session );
	assert( tech_pvt != NULL );

	channel = switch_core_session_get_channel( session );
	assert( channel != NULL );

	switch( msg->message_id )
	{
	case SWITCH_MESSAGE_INDICATE_BRIDGE:
	  if (tech_pvt->rtp_session && switch_test_flag(tech_pvt, TFLAG_TIMER)) {
	    switch_rtp_clear_flag(tech_pvt->rtp_session, SWITCH_RTP_FLAG_USE_TIMER);
	    switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "De-activate timed RTP!\n");
	  }
	  break;
	case SWITCH_MESSAGE_INDICATE_UNBRIDGE:
	  if (tech_pvt->rtp_session && switch_test_flag(tech_pvt, TFLAG_TIMER)) {
	    switch_rtp_set_flag(tech_pvt->rtp_session, SWITCH_RTP_FLAG_USE_TIMER);
	    switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Re-activate timed RTP!\n");
	  }
	  break;
	case SWITCH_MESSAGE_INDICATE_PROGRESS:
		{
			pj_status_t status;
			pjsip_tx_data *txdata;

			/* Send 183 Session Progress */
			status = pjsip_inv_answer( tech_pvt->sip_invite, 183, NULL, NULL, &txdata );
			if( status != PJ_SUCCESS ) {
				status = pjsip_inv_answer( tech_pvt->sip_invite, PJSIP_SC_NOT_ACCEPTABLE, NULL, NULL, &txdata );

				if( status == PJ_SUCCESS ) {
					add_server_header(txdata);
					pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );
				} else
					pjsip_inv_terminate( tech_pvt->sip_invite, 500, PJ_FALSE );

				switch_core_session_destroy( &session );
				return SWITCH_STATUS_GENERR;
			}
			add_server_header(txdata);
			pjsip_inv_send_msg( tech_pvt->sip_invite, txdata );
		}
		switch_channel_set_flag( channel, CF_EARLY_MEDIA );
		break;
	default:
		break;
	}

	return SWITCH_STATUS_SUCCESS;
}

/* END: pjsip callbacks */

/* BEGIN: mid-level functions */
static switch_status_t create_sdp( pj_pool_t *pool,
					struct private_object *tech_pvt,
					pjmedia_sdp_session **p_sdp,
					pjmedia_sdp_session *ref_sdp )
{
	struct pjsip_profile *profile;
	pjmedia_sdp_session *sdp;
	pjmedia_sdp_media *media;
	pjmedia_sdp_attr *attr, *rtpmap_attr;
	pjmedia_sdp_rtpmap rtpmap;
	int i;

	if( !pool || !tech_pvt )
		return SWITCH_STATUS_GENERR;

	profile = tech_pvt->profile;
	assert( profile != NULL );

	/* allocate and initialize new SDP var */
	sdp = pj_pool_zalloc( pool, sizeof(pjmedia_sdp_session) );
	if( !sdp )
		return SWITCH_STATUS_GENERR;

	sdp->origin.user = pj_str( "FreeSWITCH" );
	sdp->origin.version = sdp->origin.id = 0;
	sdp->origin.net_type = pj_str( "IN" );
	sdp->origin.addr_type = pj_str( "IP4" );
	sdp->origin.addr = pj_str( profile->local_addr );
	sdp->name = pj_str( "pjsip" );

	sdp->conn = pj_pool_zalloc( pool, sizeof(pjmedia_sdp_conn) );
	sdp->conn->net_type = pj_str( "IN" );
	sdp->conn->addr_type = pj_str( "IP4" );
	sdp->conn->addr = pj_str( profile->local_addr );

	sdp->time.start = sdp->time.stop = 0;
	sdp->attr_count = 0;

	/* add media to SDP */
	sdp->media_count = 1;
	media = pj_pool_zalloc( pool, sizeof(pjmedia_sdp_media) );
	sdp->media[0] = media;

	/* media info */
	media->desc.media = pj_str( "audio" );
	media->desc.port = tech_pvt->local_sdp_audio_port = switch_rtp_request_port();
	media->desc.port_count = 1;
	media->desc.transport = pj_str( "RTP/AVP" );

	strncpy( tech_pvt->local_sdp_audio_addr, profile->local_addr, sizeof(tech_pvt->local_sdp_audio_addr) );

	/* codec rtpmap list */
	media->desc.fmt_count = 0;
	media->attr_count = 0;

	if( !ref_sdp ) {
		for( i = 0; i < tech_pvt->num_codecs; i++ ) {
			const switch_codec_implementation_t *imp = NULL;

			imp = tech_pvt->codecs[i];

			/* add all incarnations of this codec... */
			while( imp != NULL ) {
				char tmp[10];

				/* copy codec id */
				sprintf( tmp, "%d", tech_pvt->codecs[i]->ianacode );
				pj_strdup2( pool, &media->desc.fmt[media->desc.fmt_count], tmp );

				/* other codec parameters */
				rtpmap.pt = media->desc.fmt[media->desc.fmt_count];
				rtpmap.clock_rate = imp->samples_per_second;
				rtpmap.enc_name = pj_str( tech_pvt->codecs[i]->iananame );
				rtpmap.param.slen = 0;

				/* add codec implementation to attr list */
				pjmedia_sdp_rtpmap_to_attr( pool, &rtpmap, &rtpmap_attr );
				media->attr[media->attr_count++] = rtpmap_attr;

				/* goto next implementation of this codec */
				imp = imp->next;
			}
			media->desc.fmt_count++;
		}
	}
	else {
		/* houston, we've got a reference SDP (incoming invite) 
		 * what to do next?
		 * - scan the incoming SDP for the maximum supported samplerate
		 * - build answer SDP based on the reference codec list + max samplerate
		 */
		;
	}

	/* Add sendrecv attribute */
	attr = pj_pool_zalloc( pool, sizeof(pjmedia_sdp_attr) );
	attr->name = pj_str( "sendrecv" );
	media->attr[media->attr_count++] = attr;

	/*
	 * Add support telephony event
	 */
	media->desc.fmt[media->desc.fmt_count++] = pj_str( "121" );
	/* Add rtpmap. */
	attr = pj_pool_zalloc( pool, sizeof(pjmedia_sdp_attr) );
	attr->name = pj_str( "rtpmap" );
	attr->value = pj_str( ":121 telephone-event/8000" );
	media->attr[media->attr_count++] = attr;
	/* Add fmtp */
	attr = pj_pool_zalloc( pool, sizeof(pjmedia_sdp_attr) );
	attr->name = pj_str( "fmtp" );
	attr->value = pj_str( ":121 0-15" );
	media->attr[media->attr_count++] = attr;

	/* all done */
	*p_sdp = sdp;
	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t start_rtp( struct private_object *tech_pvt )
{
	int rate, ms;
	const char *err;
	switch_rtp_flag_t flags = 0;
	switch_channel_t *channel;

	assert( tech_pvt != NULL );

	channel = switch_core_session_get_channel( tech_pvt->session );
	assert( channel != NULL );

	if( tech_pvt->rtp_session ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "RTP already initialized %s\n", tech_pvt->call_id );
		return SWITCH_STATUS_SUCCESS;
	}

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Starting RTP for call %s\n", tech_pvt->call_id );

	rate = tech_pvt->read_codec.implementation->bits_per_second;
	ms   = tech_pvt->read_codec.implementation->microseconds_per_frame;

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "RTP parameters for call %s: src (%s:%d) -> dst (%s:%d), using codec %s @ %d b/s, framelen %d ms\n",
					tech_pvt->call_id,
					tech_pvt->local_sdp_audio_addr,
					tech_pvt->local_sdp_audio_port,
					tech_pvt->remote_sdp_audio_addr,
					tech_pvt->remote_sdp_audio_port,
					tech_pvt->read_codec.implementation->iananame,
					rate,
					ms );

	flags = SWITCH_RTP_FLAG_RAW_WRITE | SWITCH_RTP_FLAG_USE_TIMER;

	tech_pvt->rtp_session = switch_rtp_new( tech_pvt->local_sdp_audio_addr,
											tech_pvt->local_sdp_audio_port,
											tech_pvt->remote_sdp_audio_addr,
											tech_pvt->remote_sdp_audio_port,
											tech_pvt->read_codec.implementation->ianacode,
											0,
											tech_pvt->read_codec.implementation->microseconds_per_frame,
											flags,
											NULL,
											&err, switch_core_session_get_pool( tech_pvt->session ) );

	if( !tech_pvt->rtp_session ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "RTP ERROR [%s]\n", err);
		switch_channel_hangup( channel, SWITCH_CAUSE_DESTINATION_OUT_OF_ORDER );
		switch_set_flag_locked( tech_pvt, TFLAG_BYE );
		return SWITCH_STATUS_FALSE;
	}

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Starting RTP done\n" );

	switch_set_flag_locked( tech_pvt, TFLAG_IO);

	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t stop_rtp( struct private_object *tech_pvt )
{
	int loop = 0;

	if( !tech_pvt->rtp_session )
		return SWITCH_STATUS_SUCCESS;

	while( loop < 10 &&
	       ( switch_test_flag( tech_pvt, TFLAG_READING ) ||
		 switch_test_flag( tech_pvt, TFLAG_WRITING ) ) )
	{
		switch_yield( 1000 );
		loop++;
	}

	switch_mutex_lock(tech_pvt->flag_mutex);
	switch_rtp_destroy( &tech_pvt->rtp_session );
	tech_pvt->rtp_session = NULL;
	switch_mutex_unlock(tech_pvt->flag_mutex);
	
	return SWITCH_STATUS_SUCCESS;
}

static switch_status_t config_pjsip( void )
{
	char *cfgname = MOD_PJSIP_CFG;
	switch_xml_t config, xml, xmlprofile, param;
	switch_status_t result = SWITCH_STATUS_SUCCESS;

	if( !( xml = switch_xml_open_cfg( cfgname, &config, NULL ) )) {
                switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "Could not open config %s\n", cfgname );
                return SWITCH_STATUS_TERM;
	}

	for( xmlprofile = switch_xml_child( config, "profile" ); xmlprofile; xmlprofile = xmlprofile->next )
	{
		struct pjsip_profile *profile;
		char *profile_name = (char *) switch_xml_attr_soft( xmlprofile, "name" );

		if( !profile_name ) {
			profile_name = "noname";
		}

		if( !(profile = switch_core_alloc( module_pool, sizeof(struct pjsip_profile))) ) {
	                switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "Failed to allocate new profile %s\n", profile_name );
        	        result = SWITCH_STATUS_TERM;
			goto out;
		}
		memset( profile, 0, sizeof( struct pjsip_profile ) );

		switch_copy_string( profile->name, profile_name, sizeof( profile->name ) );

		profile->sip_addr.sin_family = PJ_AF_INET;
		profile->sip_addr.sin_port = pj_htons( DEFAULT_PORT );

		for( param = switch_xml_child( xmlprofile, "param" ); param; param = param->next )
		{
			char *var = (char *) switch_xml_attr_soft( param, "name" );
			char *val = (char *) switch_xml_attr_soft( param, "value" );

			if ( !strcmp( var, "max-calls" ) ) {
				profile->max_calls = atoi( val );
			}
			else if ( !strcmp( var, "ip" ) ) {
				pj_str_t tmp;

				if( !pj_inet_aton( pj_cstr( &tmp, val ), &profile->sip_addr.sin_addr ) ) {
	 				switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Invalid address: %s\n", val );
					result = SWITCH_STATUS_TERM;
					goto out;
				}
				set_param( profile->sip_ip, val );
			}
			else if ( !strcmp( var, "port" ) ) {
				profile->sip_port = atoi( val );
				profile->sip_addr.sin_port = pj_htons( profile->sip_port );
			}
			else if ( !strcmp( var, "dialplan" ) ) {
				set_param( profile->dialplan, val );
			}
			else if ( !strcmp( var, "codec-prefs" ) ) {
				set_param( profile->codec_string, val );
				profile->codec_order_last = switch_separate_string( profile->codec_string, ',', profile->codec_order, SWITCH_MAX_CODECS );
			}
			else if ( !strcmp( var, "transport" ) ) {
				char *tmp[MAX_TRANSPORTS];
				int max, n;

				max = switch_separate_string( val, ',', tmp, MAX_TRANSPORTS );
				for( n = 0; n < max; n++ ) {
					char *tp = tmp[n];

					if( !strcmp( tp, "tcp" ) ) {
						switch_set_flag( profile, PFLAG_TRANSPORT_TCP );
					}
					else if( !strcmp( tp, "udp" ) ) {
						switch_set_flag( profile, PFLAG_TRANSPORT_UDP );
					}
					else if( !strcmp( tp, "tls" ) ) {
						switch_set_flag( profile, PFLAG_TRANSPORT_TLS );
					}
					else {
						switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "Unknown transport: %s\n", tp );
						result = SWITCH_STATUS_TERM;
						goto out;
					}
				}
			}
			else {
				switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "Unknown option %s (ignored)\n", var );
			}
		}		

		/* set default udp transport if none given */
		if( !switch_test_flag( profile, PFLAG_TRANSPORT_UDP ) &&
		    !switch_test_flag( profile, PFLAG_TRANSPORT_TCP ) &&
		    !switch_test_flag( profile, PFLAG_TRANSPORT_TLS ) )
		{
			switch_set_flag( profile, PFLAG_TRANSPORT_UDP );
		}

		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "Adding profile %s\n", profile->name );
		switch_core_hash_insert( globals.profiles_hash, profile->name, profile );
	}

out:
	switch_xml_free( xml );

	return result;
}
/* END: mid-level functions */


/* BEGIN: module managment */

SWITCH_MOD_DECLARE(switch_status_t) switch_module_shutdown(void)
{
        if ( globals.running ) {
                globals.running = -1;
                while ( globals.running ) {
                        switch_yield( 100000 );
                }
        }
        return SWITCH_STATUS_SUCCESS;
}

SWITCH_MOD_DECLARE(switch_status_t) switch_module_load(const switch_loadable_module_interface_t **module_interface, char *filename)
{
        /* NOTE:  **interface is **_interface because the common lib redefines interface to struct in some situations */
        if (switch_core_new_memory_pool( &module_pool ) != SWITCH_STATUS_SUCCESS ) {
                switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_CRIT, "OH OH no pool\n" );
                return SWITCH_STATUS_TERM;
        }

        switch_core_hash_init( &globals.profiles_hash, module_pool );
        switch_core_hash_init( &globals.transport_hash, module_pool );

        /* connect my internal structure to the blank pointer passed to me */
        *module_interface = &pjsip_module_interface;

        /* indicate that the module should continue to be loaded */
        return SWITCH_STATUS_SUCCESS;
}


/*****************************************************************************
 * This is a very simple PJSIP module, whose sole purpose is to display
 * incoming and outgoing messages to log. This module will have priority
 * higher than transport layer, which means:
 *
 *  - incoming messages will come to this module first before reaching
 *    transaction layer.
 *
 *  - outgoing messages will come to this module last, after the message
 *    has been 'printed' to contiguous buffer by transport layer and
 *    appropriate transport instance has been decided for this message.
 *
 */

/* Notification on incoming messages */
static pj_bool_t logging_on_rx_msg(pjsip_rx_data *rdata)
{
    PJ_LOG(4,("mod_pjsip", "RX %d bytes %s from %s %s:%d:\n"
			 "%.*s\n"
			 "--end msg--",
			 rdata->msg_info.len,
			 pjsip_rx_data_get_info(rdata),
			 rdata->tp_info.transport->type_name,
			 rdata->pkt_info.src_name,
			 rdata->pkt_info.src_port,
			 (int)rdata->msg_info.len,
			 rdata->msg_info.msg_buf));
    
    /* Always return false, otherwise messages will not get processed! */
    return PJ_FALSE;
}

/* Notification on outgoing messages */
static pj_status_t logging_on_tx_msg(pjsip_tx_data *tdata)
{
    
    /* Important note:
     *	tp_info field is only valid after outgoing messages has passed
     *	transport layer. So don't try to access tp_info when the module
     *	has lower priority than transport layer.
     */

    PJ_LOG(4,("mod_pjsip", "TX %d bytes %s to %s %s:%d:\n"
			 "%.*s\n"
			 "--end msg--",
			 (tdata->buf.cur - tdata->buf.start),
			 pjsip_tx_data_get_info(tdata),
			 tdata->tp_info.transport->type_name,
			 tdata->tp_info.dst_name,
			 tdata->tp_info.dst_port,
			 (int)(tdata->buf.cur - tdata->buf.start),
			 tdata->buf.start));

    /* Always return success, otherwise message will not get sent! */
    return PJ_SUCCESS;
}

/* The module instance. */
static pjsip_module pjsua_msg_logger = 
{
    NULL, NULL,				/* prev, next.		*/
    { "mod-pjsua-log", 13 },		/* Name.		*/
    -1,					/* Id			*/
    PJSIP_MOD_PRIORITY_TRANSPORT_LAYER-1,/* Priority	        */
    NULL,				/* load()		*/
    NULL,				/* start()		*/
    NULL,				/* stop()		*/
    NULL,				/* unload()		*/
    &logging_on_rx_msg,			/* on_rx_request()	*/
    &logging_on_rx_msg,			/* on_rx_response()	*/
    &logging_on_tx_msg,			/* on_tx_request.	*/
    &logging_on_tx_msg,			/* on_tx_response()	*/
    NULL,				/* on_tsx_state()	*/

};






SWITCH_MOD_DECLARE(switch_status_t) switch_module_runtime(void)
{
	switch_hash_index_t *p_entry;
	pj_status_t status;

	/* init pjlib */
	status = pj_init();
	if ( status != PJ_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize pjlib\n" );
		return SWITCH_STATUS_TERM;
	}

	pj_log_set_log_func(log_func);
	pj_log_set_level(5);

	/* next: pjlib_util */
	status = pjlib_util_init();
	if ( status != PJ_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize pjlib_util\n" );
		return SWITCH_STATUS_TERM;
	}

	/* initialize pjlib pool factory */
	pj_caching_pool_init( &cp, &pj_pool_factory_default_policy, 0 );

	/* global pool */
	pjsip_pool = pj_pool_create( &cp.factory, "pjsip-pool", 1024, 1024, NULL );
	if( !pjsip_pool ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize global pool\n" );
		return SWITCH_STATUS_TERM;
	}

	/* Initialize server name */
	pj_strdup2_with_null(pjsip_pool, &FREESWITCH_SERVER, "FreeSWITCH");

	if( config_pjsip( ) != SWITCH_STATUS_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to load configuration\n" );
		return SWITCH_STATUS_TERM;
	}

	/* create endpoint */
	status = pjsip_endpt_create( &cp.factory, "FreeSWITCH", &globals.sip_endpt );
	if ( status != PJ_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize sip endpoint\n" );
		return SWITCH_STATUS_TERM;
	}

	pj_memset( &invite_cb, 0, sizeof(invite_cb) );
	invite_cb.on_state_changed = &call_on_state_changed;
	invite_cb.on_new_session = &call_on_forked;
	invite_cb.on_media_update = &call_on_media_update;
	invite_cb.on_tsx_state_changed = &call_on_tsx_state_changed;

	/*
	 * Register logger module.
	 */
	pjsip_endpt_register_module(globals.sip_endpt, &pjsua_msg_logger);

	/*
	 * 
	 */
	for( p_entry = switch_hash_first( module_pool, globals.profiles_hash ); p_entry; p_entry = switch_hash_next( p_entry ) )
	{
		struct pjsip_profile *profile;
		void *val;

		switch_hash_this( p_entry, NULL, NULL, &val );
		profile = val;

		if( switch_test_flag( profile, PFLAG_TRANSPORT_UDP ) ) {
			pjsip_transport *transport;

			/*
			 * Standard UDP SIP Transport 
			 */
			status = pjsip_udp_transport_start( globals.sip_endpt, &profile->sip_addr, NULL, 1, &transport );
			if ( status != PJ_SUCCESS ) {
	 			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create UDP transport: %s [%s:%d]\n", profile->name, profile->sip_ip, profile->sip_port );
				return SWITCH_STATUS_TERM;
			}

			/* add transports name to hash */
			switch_core_hash_insert( globals.transport_hash, transport->obj_name, profile );
		}
#if 0
/* pjsip_tpfactory doesn't have an obj_name,
   will have to find another solution to do the
   incoming request -> profile lookup in on_incoming_call() */

		if( switch_test_flag( profile, PFLAG_TRANSPORT_TCP ) ) {
			pjsip_tpfactory *transport;

			/*
			 * TCP SIP Transport 
			 */
			status = pjsip_tcp_transport_start( globals.sip_endpt, &profile->sip_addr, 1, &transport );
			if ( status != PJ_SUCCESS ) {
	 			switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create TCP transport: %s [%s:%d]\n", profile->name, profile->sip_ip, profile->sip_port );
				return SWITCH_STATUS_TERM;
			}

			/* add transports name to hash */
			switch_core_hash_insert( globals.transport_hash, transport->obj_name, profile );
		}
#endif
		/* */
		{
			char *addr;
			char local_uri[64];

			/* if bind addr is 0.0.0.0 use local hostname to get the primary ip */
			if( !profile->sip_addr.sin_addr.s_addr ) {
				const pj_str_t *hostname;
				pj_sockaddr_in tmp_addr;

				hostname = pj_gethostname();
				pj_sockaddr_in_init( &tmp_addr, hostname, 0 );
				addr = pj_inet_ntoa( tmp_addr.sin_addr );
			} else
				addr = profile->sip_ip;

			pj_ansi_strcpy( profile->local_addr, addr );
			pj_ansi_sprintf( local_uri, "sip:%s:%d", addr, profile->sip_port );
			pj_strdup2( pjsip_pool, &profile->local_uri, local_uri );
			pj_strdup2( pjsip_pool, &profile->local_contact, local_uri );
		}

		/* init call hash */
        	switch_core_hash_init( &profile->call_hash, module_pool );
	}

	/* init transaction handling */
	status = pjsip_tsx_layer_init_module( globals.sip_endpt );
	if ( status != PJ_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize transaction handling\n" );
		return SWITCH_STATUS_TERM;
	}

	/* init UA dialog handling */
	status = pjsip_ua_init_module( globals.sip_endpt, NULL );
	if ( status != PJ_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to initialize UA dialog handling\n" );
		return SWITCH_STATUS_TERM;
	}

	/* initialize invite session module */
	status = pjsip_inv_usage_init( globals.sip_endpt, &invite_cb );
	if ( status != PJ_SUCCESS ) {
 		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to register invite session module to pjsip core\n" );
		return SWITCH_STATUS_TERM;
	}

	pj_memset( &globals.mod_app, 0, sizeof(globals.mod_app) );
	globals.mod_app.name = pj_str( "mod_pjsip" );			/* Name.		*/
	globals.mod_app.id = -1;					/* Id			*/
	globals.mod_app.priority = PJSIP_MOD_PRIORITY_APPLICATION;	/* Priority		*/
	globals.mod_app.on_rx_request = &on_rx_request;			/* on_rx_request()	*/

	/* register application to pjsip core */
	status = pjsip_endpt_register_module( globals.sip_endpt, &globals.mod_app );
	if ( status != PJ_SUCCESS ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to register mod_pjsip application to pjsip core\n" );
		return SWITCH_STATUS_TERM;
	}

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "Monitor thread running\n" );

	/* main loop */
	globals.running = 1;
	while ( globals.running > 0 ) {
		pj_time_val timeout = { 1, 0 };
		pjsip_endpt_handle_events( globals.sip_endpt, &timeout );
	}
	globals.running = 0;

	switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "Monitor thread exiting\n" );

	return SWITCH_STATUS_TERM;
}
